/* Pango
 * pangogp-to-pdf.c: Example program to view a UTF-8 encoding file
 *                   using Pango to render result.
 *
 * Copyright (C) 1999 Red Hat Software
 * Copyright (C) 2001, 2003 Sun Microsystems
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define BUFSIZE 1024
#define MALLOCSIZE 1024

#include <pango/pango.h>
#include <pango/pangogp.h>
#include <libgnomeprint/gnome-print-job.h>
#include <libgnomeprint/gnome-print-paper.h>

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

char *prog_name = NULL;
typedef struct _Paragraph Paragraph;

/* Structure representing a paragraph
 */
struct _Paragraph {
  char *text;
  int length;
  int height;   /* Height, in pixels */
  PangoLayout *layout;
};

GList *paragraphs;

static PangoFontDescription *font_description;

PangoContext *context;

/* Read an entire file into a string
 */
static char *
read_file (char *name)
{
  GString *inbuf;
  FILE *file;
  char *text;
  char buffer[BUFSIZE];

  file = fopen (name, "r");
  if (!file)
    {
      fprintf (stderr, "%s: Cannot open %s\n", g_get_prgname (), name);
      return NULL;
    }

  inbuf = g_string_new (NULL);
  while (1)
    {
      char *bp = fgets (buffer, BUFSIZE-1, file);
      if (ferror (file))
	{
	  fprintf(stderr, "%s: Error reading %s\n", g_get_prgname (), name);
	  g_string_free (inbuf, TRUE);
	  return NULL;
	}
      else if (bp == NULL)
	break;

      g_string_append (inbuf, buffer);
    }

  fclose (file);

  text = inbuf->str;
  g_string_free (inbuf, FALSE);

  return text;
}

/* Take a UTF8 string and break it into paragraphs on \n characters
 */
static GList *
split_paragraphs (char *text)
{
  char *p = text;
  char *next;
  gunichar wc;
  GList *result = NULL;
  char *last_para = text;
  
  while (*p)
    {
      wc = g_utf8_get_char (p);
      next = g_utf8_next_char (p);
      if (wc == (gunichar)-1)
	{
	  fprintf (stderr, "%s: Invalid character in input\n", g_get_prgname ());
	  wc = 0;
	}
      if (!*p || !wc || wc == '\n')
	{
	  Paragraph *para = g_new (Paragraph, 1);
	  para->text = last_para;
	  para->length = p - last_para;
	  para->layout = pango_layout_new (context);
	  pango_layout_set_text (para->layout, para->text, para->length);
	  para->height = 0;

	  last_para = next;
	    
	  result = g_list_prepend (result, para);
	}
      if (!wc) /* incomplete character at end */
	break;
      p = next;
    }

  return g_list_reverse (result);
}

int main(int argc, char *argv[])
{
  char *init_family = "sans";
  int init_scale = 24;
  PangoDirection init_dir = PANGO_DIRECTION_LTR;
  char *text;
  int y_start = 100, x_start = 100;
  int init_size = 24;
  int argp;
  char *prog_name = g_path_get_basename (argv[0]);
  GnomePrintContext *gpc;
  GnomePrintJob *job;
  GnomePrintConfig *config;
  gchar *test;
  gchar *output;
  gchar *init_format = "pdf";
  gchar *init_paper = "a4";
  gint ret = 0;
  gdouble width, height;
  const GnomePrintUnit *ps_unit = GNOME_PRINT_PS_UNIT;
  gdouble ps_unit_multiplier;

  g_type_init();
  
  job = gnome_print_job_new (NULL);
  if (!job)
    fprintf (stderr, "Print job new");
  gpc = gnome_print_job_get_context (job);
  if (!gpc)
    fprintf (stderr, "Print job get context");
  config = gnome_print_job_get_config (job);
  if (!config)
    fprintf (stderr, "Print job get config");

  /* Parse command line */
  argp=1;
  while(argp < argc && argv[argp][0] == '-')
    {
      char *opt = argv[argp++];
      if (strcmp(opt, "--help") == 0)
	{
	  printf("%s - An example viewer for the Pango GNOME Print (pangogp) backend\n"
		 "\n"
		 "Syntax:\n"
		 "    %s [--family f] [--format {pdf|ps}] [--size s] file\n"
		 "\n"
		 "Options:\n"
		 "    --family f            Set the initial family. Default is '%s'.\n"
		 "    --format {pdf|ps}     Set the output format. Default is '%s'.\n"
		 /*"    --paper {A4|USLetter} Set the output paper size. Default is '%s'.\n"*/
		 "    --size s              Set the initial font size. Default is %d\n"
		 "    --rtl                 Set base dir to RTL. Default is LTR.\n",
		 prog_name, prog_name, init_family, init_format, /*init_paper,*/
		 init_scale);
	  exit(0);
	}
      if (strcmp (opt, "--family") == 0)
	{
	  init_family = argv[argp++];
	  continue;
	}
      if (strcmp (opt, "--format") == 0)
	{
	  init_format = argv[argp++];
	  continue;
	}
#if 0
      if (strcmp (opt, "--paper") == 0)
	{
	  init_paper = argv[argp++];

	  /*if ((strcmp (init_paper, "A4") == 0) ||
	    (strcmp (init_paper, "USLetter") == 0))*/
	    {
	      const GnomePrintPaper *paper;

	      paper = gnome_print_paper_get_by_name (init_paper);

	      if (paper != NULL)
		{
		  gnome_print_config_set (config,
					  GNOME_PRINT_KEY_PAPER_SIZE,
					  paper->name);
		  gnome_print_config_set_length (config,
						 GNOME_PRINT_KEY_PAPER_WIDTH,
						 paper->width,
						 GNOME_PRINT_PS_UNIT);
		  gnome_print_config_set_length (config,
						 GNOME_PRINT_KEY_PAPER_HEIGHT,
						 paper->height,
						 GNOME_PRINT_PS_UNIT);
		 }
	      else
		{
		  fprintf (stderr,
			   "gnome_print_config_set %s-%s\n",
			   GNOME_PRINT_KEY_PAPER_SIZE,
			   init_paper);
		}
	    }
	    /*
	  else
	    {
	      fprintf (stderr, "Unknown paper size: '%s'\n", init_paper);
	      exit (1);
	    }
	    */
	  continue;
	}
#endif
      if (strcmp (opt, "--size") == 0)
	{
	  init_size = atoi(argv[argp++]);
	  continue;
	}
      if (strcmp (opt, "--rtl") == 0)
	{
	  init_dir = PANGO_DIRECTION_RTL;
	  continue;
	}
      fprintf (stderr, "Unknown option %s!\n", opt);
      exit (1);
    }

  if (argp + 1 != argc && argp + 2 != argc)
    {
      fprintf (stderr, "Usage: %s [options] FILE [OUTFILE]\n", prog_name);
      exit (1);
    }

  /* Create the list of paragraphs from the supplied file
   */
  text = read_file (argv[argp++]);
  if (!text)
    exit(1);

  if (argp < argc)
    output = argv[argp++];

  if (strcmp (init_format, "pdf") == 0)
    {
      if (!gnome_print_config_set (config, "Printer", "PDF"))
	fprintf (stderr, "gnome_print_config_set Printer-PDF");
      test = gnome_print_config_get (config, "Printer");
      if (!test)
	fprintf (stderr, "gnome_print_config_get Printer returned NULL\n");
      if (strcmp (test, "PDF") != 0)
	fprintf (stderr, "Could not set printer to PDF.\n");
      g_free (test);		

      /*
       * libgnomeprint-2.2.1.2 PDF backend reports page sizes in points.
       */
      ps_unit_multiplier = 1.0;

      output = "pangogp.pdf";
    }
  else if (strcmp (init_format, "ps") == 0)
    {
      if (!gnome_print_config_set (config, "Printer", "GENERIC"))
	fprintf (stderr, "gnome_print_config_set Printer-GENERIC");

      /*
       * libgnomeprint-2.2.1.2 PS backend reports page sizes in millimetres.
       */
      ps_unit_multiplier = 72 / 25.4;

      output = "pangogp.ps";
    }
  else
    {
      fprintf (stderr, "Unknown output format: '%s'\n", init_format);
      exit (1);
    }

  ret += gnome_print_beginpage (gpc, "1");

  gnome_print_config_get_length (config,
				 GNOME_PRINT_KEY_PAPER_WIDTH,
				 &width,
				 &ps_unit);
  gnome_print_config_get_length (config,
				 GNOME_PRINT_KEY_PAPER_HEIGHT,
				 &height,
				 &ps_unit);

  g_message ("width: %g; height: %g",
	     width * ps_unit_multiplier,
	     height * ps_unit_multiplier);

  context = pango_gp_get_context (gpc);

  paragraphs = split_paragraphs (text);

  pango_context_set_language (context, pango_language_from_string ("en_US"));
  pango_context_set_base_dir (context, init_dir);

  font_description = pango_font_description_new ();
  pango_font_description_set_family (font_description, g_strdup (init_family));
  pango_font_description_set_style (font_description, PANGO_STYLE_NORMAL);
  pango_font_description_set_variant (font_description, PANGO_VARIANT_NORMAL);
  pango_font_description_set_weight (font_description, PANGO_WEIGHT_NORMAL);
  pango_font_description_set_stretch (font_description, PANGO_STRETCH_NORMAL);
  pango_font_description_set_size (font_description, init_scale * PANGO_SCALE);

  pango_context_set_font_description (context, font_description);

  /* Write first paragraph to output file */
  {
      Paragraph *para = paragraphs->data;
      PangoDirection base_dir = pango_context_get_base_dir (context);
      PangoRectangle logical_rect;

      pango_layout_set_alignment (para->layout,
				  base_dir == PANGO_DIRECTION_LTR ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
      pango_layout_set_width (para->layout,
			      ((width * ps_unit_multiplier) - 2 * 72) * PANGO_SCALE);

      pango_layout_get_extents (para->layout, NULL, &logical_rect);
      pango_gp_render_layout (context,
			      para->layout,
			      72 * PANGO_SCALE,
			      ((height * ps_unit_multiplier) - 72) * PANGO_SCALE);
  }

  ret += gnome_print_showpage (gpc);

  if (gnome_print_job_print_to_file (job, output) != GNOME_PRINT_OK)
    fprintf (stderr, "print_job_print_to_file");

  if (ret > 0)
    fprintf (stderr, "drawing functions");
  if (gnome_print_job_close (job) != GNOME_PRINT_OK)
    fprintf (stderr, "gnome_print_job_close");
  if (gnome_print_job_print (job) != GNOME_PRINT_OK)
    fprintf (stderr, "gnome_print_job_print");

  g_object_unref (config);
  g_object_unref (gpc);
  g_object_unref (job);

  return ret;
}
